# -*- coding: utf-8 -*-
"""This package is a framework for creating large-scale agent-based geographical models.

Submodules
==========


"""

__title__ = "honeybees"
__version__ = "1.0.6"
__email__ = "j.a.debruijn at outlook com"
__status__ = "development"
